package java.awt.geom;

import java.io.Serializable;
import java.awt.Rectangle;
import java.awt.Shape;

public abstract class QuadCurve2D implements Shape, Cloneable
{
    private static final int BELOW = -2;
    private static final int LOWEDGE = -1;
    private static final int INSIDE = 0;
    private static final int HIGHEDGE = 1;
    private static final int ABOVE = 2;
    
    protected QuadCurve2D() {
    }
    
    public abstract double getX1();
    
    public abstract double getY1();
    
    public abstract Point2D getP1();
    
    public abstract double getCtrlX();
    
    public abstract double getCtrlY();
    
    public abstract Point2D getCtrlPt();
    
    public abstract double getX2();
    
    public abstract double getY2();
    
    public abstract Point2D getP2();
    
    public abstract void setCurve(double p0, double p1, double p2, double p3, double p4, double p5);
    
    public void setCurve(double[] array, int n) {
        this.setCurve(array[n + 0], array[n + 1], array[n + 2], array[n + 3], array[n + 4], array[n + 5]);
    }
    
    public void setCurve(Point2D point2D, Point2D point2D2, Point2D point2D3) {
        this.setCurve(point2D.getX(), point2D.getY(), point2D2.getX(), point2D2.getY(), point2D3.getX(), point2D3.getY());
    }
    
    public void setCurve(Point2D[] array, int n) {
        this.setCurve(array[n + 0].getX(), array[n + 0].getY(), array[n + 1].getX(), array[n + 1].getY(), array[n + 2].getX(), array[n + 2].getY());
    }
    
    public void setCurve(QuadCurve2D quadCurve2D) {
        this.setCurve(quadCurve2D.getX1(), quadCurve2D.getY1(), quadCurve2D.getCtrlX(), quadCurve2D.getCtrlY(), quadCurve2D.getX2(), quadCurve2D.getY2());
    }
    
    public static double getFlatnessSq(double n, double n2, double n3, double n4, double n5, double n6) {
        return Line2D.ptSegDistSq(n, n2, n5, n6, n3, n4);
    }
    
    public static double getFlatness(double n, double n2, double n3, double n4, double n5, double n6) {
        return Line2D.ptSegDist(n, n2, n5, n6, n3, n4);
    }
    
    public static double getFlatnessSq(double[] array, int n) {
        return Line2D.ptSegDistSq(array[n + 0], array[n + 1], array[n + 4], array[n + 5], array[n + 2], array[n + 3]);
    }
    
    public static double getFlatness(double[] array, int n) {
        return Line2D.ptSegDist(array[n + 0], array[n + 1], array[n + 4], array[n + 5], array[n + 2], array[n + 3]);
    }
    
    public double getFlatnessSq() {
        return Line2D.ptSegDistSq(this.getX1(), this.getY1(), this.getX2(), this.getY2(), this.getCtrlX(), this.getCtrlY());
    }
    
    public double getFlatness() {
        return Line2D.ptSegDist(this.getX1(), this.getY1(), this.getX2(), this.getY2(), this.getCtrlX(), this.getCtrlY());
    }
    
    public void subdivide(QuadCurve2D quadCurve2D, QuadCurve2D quadCurve2D2) {
        subdivide(this, quadCurve2D, quadCurve2D2);
    }
    
    public static void subdivide(QuadCurve2D quadCurve2D, QuadCurve2D quadCurve2D2, QuadCurve2D quadCurve2D3) {
        double x1 = quadCurve2D.getX1();
        double y1 = quadCurve2D.getY1();
        double ctrlX = quadCurve2D.getCtrlX();
        double ctrlY = quadCurve2D.getCtrlY();
        double x2 = quadCurve2D.getX2();
        double y2 = quadCurve2D.getY2();
        double n = (x1 + ctrlX) / 2.0;
        double n2 = (y1 + ctrlY) / 2.0;
        double n3 = (x2 + ctrlX) / 2.0;
        double n4 = (y2 + ctrlY) / 2.0;
        double n5 = (n + n3) / 2.0;
        double n6 = (n2 + n4) / 2.0;
        if (quadCurve2D2 != null) {
            quadCurve2D2.setCurve(x1, y1, n, n2, n5, n6);
        }
        if (quadCurve2D3 != null) {
            quadCurve2D3.setCurve(n5, n6, n3, n4, x2, y2);
        }
    }
    
    public static void subdivide(double[] array, int n, double[] array2, int n2, double[] array3, int n3) {
        double n4 = array[n + 0];
        double n5 = array[n + 1];
        double n6 = array[n + 2];
        double n7 = array[n + 3];
        double n8 = array[n + 4];
        double n9 = array[n + 5];
        if (array2 != null) {
            array2[n2 + 0] = n4;
            array2[n2 + 1] = n5;
        }
        if (array3 != null) {
            array3[n3 + 4] = n8;
            array3[n3 + 5] = n9;
        }
        double n10 = (n4 + n6) / 2.0;
        double n11 = (n5 + n7) / 2.0;
        double n12 = (n8 + n6) / 2.0;
        double n13 = (n9 + n7) / 2.0;
        double n14 = (n10 + n12) / 2.0;
        double n15 = (n11 + n13) / 2.0;
        if (array2 != null) {
            array2[n2 + 2] = n10;
            array2[n2 + 3] = n11;
            array2[n2 + 4] = n14;
            array2[n2 + 5] = n15;
        }
        if (array3 != null) {
            array3[n3 + 0] = n14;
            array3[n3 + 1] = n15;
            array3[n3 + 2] = n12;
            array3[n3 + 3] = n13;
        }
    }
    
    public static int solveQuadratic(double[] array) {
        return solveQuadratic(array, array);
    }
    
    public static int solveQuadratic(double[] array, double[] array2) {
        double n = array[2];
        double n2 = array[1];
        double n3 = array[0];
        int n4 = 0;
        if (n == 0.0) {
            if (n2 == 0.0) {
                return -1;
            }
            array2[n4++] = -n3 / n2;
        }
        else {
            double a = n2 * n2 - 4.0 * n * n3;
            if (a < 0.0) {
                return 0;
            }
            double sqrt = Math.sqrt(a);
            if (n2 < 0.0) {
                sqrt = -sqrt;
            }
            double n5 = (n2 + sqrt) / -2.0;
            array2[n4++] = n5 / n;
            if (n5 != 0.0) {
                array2[n4++] = n3 / n5;
            }
        }
        return n4;
    }
    
    @Override
    public boolean contains(double n, double n2) {
        double x1 = this.getX1();
        double y1 = this.getY1();
        double ctrlX = this.getCtrlX();
        double ctrlY = this.getCtrlY();
        double x2 = this.getX2();
        double y2 = this.getY2();
        double n3 = x1 - 2.0 * ctrlX + x2;
        double n4 = y1 - 2.0 * ctrlY + y2;
        double n5 = n - x1;
        double n6 = n2 - y1;
        double n7 = x2 - x1;
        double n8 = y2 - y1;
        double n9 = (n5 * n4 - n6 * n3) / (n7 * n4 - n8 * n3);
        if (n9 < 0.0 || n9 > 1.0 || n9 != n9) {
            return false;
        }
        double n10 = n3 * n9 * n9 + 2.0 * (ctrlX - x1) * n9 + x1;
        double n11 = n4 * n9 * n9 + 2.0 * (ctrlY - y1) * n9 + y1;
        double n12 = n7 * n9 + x1;
        double n13 = n8 * n9 + y1;
        return (n >= n10 && n < n12) || (n >= n12 && n < n10) || (n2 >= n11 && n2 < n13) || (n2 >= n13 && n2 < n11);
    }
    
    @Override
    public boolean contains(Point2D point2D) {
        return this.contains(point2D.getX(), point2D.getY());
    }
    
    private static void fillEqn(double[] array, double n, double n2, double n3, double n4) {
        array[0] = n2 - n;
        array[1] = n3 + n3 - n2 - n2;
        array[2] = n2 - n3 - n3 + n4;
    }
    
    private static int evalQuadratic(double[] array, int n, boolean b, boolean b2, double[] array2, double n2, double n3, double n4) {
        int n5 = 0;
        for (double n6 : array) {
            if (!(b ? (n6 < 0.0) : (n6 <= 0.0)) && !(b2 ? (n6 > 1.0) : (n6 >= 1.0)) && (array2 == null || array2[1] + 2.0 * array2[2] * n6 != 0.0)) {
                double n7 = 1.0 - n6;
                array[n5++] = n2 * n7 * n7 + 2.0 * n3 * n6 * n7 + n4 * n6 * n6;
            }
        }
        return n5;
    }
    
    private static int getTag(double n, double n2, double n3) {
        if (n <= n2) {
            return (n < n2) ? -2 : -1;
        }
        if (n >= n3) {
            return (n > n3) ? 2 : 1;
        }
        return 0;
    }
    
    private static boolean inwards(int n, int n2, int n3) {
        switch (n) {
            case -1: {
                return n2 >= 0 || n3 >= 0;
            }
            case 0: {
                return true;
            }
            case 1: {
                return n2 <= 0 || n3 <= 0;
            }
            default: {
                return false;
            }
        }
    }
    
    @Override
    public boolean intersects(double n, double n2, double n3, double n4) {
        if (n3 <= 0.0 || n4 <= 0.0) {
            return false;
        }
        double x1 = this.getX1();
        double y1 = this.getY1();
        int n5 = getTag(x1, n, n + n3);
        int tag = getTag(y1, n2, n2 + n4);
        if (n5 == 0 && tag == 0) {
            return true;
        }
        double x2 = this.getX2();
        double y2 = this.getY2();
        int n6 = getTag(x2, n, n + n3);
        int tag2 = getTag(y2, n2, n2 + n4);
        if (n6 == 0 && tag2 == 0) {
            return true;
        }
        double ctrlX = this.getCtrlX();
        double ctrlY = this.getCtrlY();
        int tag3 = getTag(ctrlX, n, n + n3);
        int tag4 = getTag(ctrlY, n2, n2 + n4);
        if (n5 < 0 && n6 < 0 && tag3 < 0) {
            return false;
        }
        if (tag < 0 && tag2 < 0 && tag4 < 0) {
            return false;
        }
        if (n5 > 0 && n6 > 0 && tag3 > 0) {
            return false;
        }
        if (tag > 0 && tag2 > 0 && tag4 > 0) {
            return false;
        }
        if (inwards(n5, n6, tag3) && inwards(tag, tag2, tag4)) {
            return true;
        }
        if (inwards(n6, n5, tag3) && inwards(tag2, tag, tag4)) {
            return true;
        }
        int n7 = n5 * n6 <= 0;
        int n8 = tag * tag2 <= 0;
        if (n5 == 0 && n6 == 0 && n8) {
            return true;
        }
        if (tag == 0 && tag2 == 0 && n7) {
            return true;
        }
        double[] array = new double[3];
        double[] array2 = new double[3];
        if (n8 == 0) {
            fillEqn(array, (tag < 0) ? n2 : (n2 + n4), y1, ctrlY, y2);
            return solveQuadratic(array, array2) == 2 && evalQuadratic(array2, 2, true, true, null, x1, ctrlX, x2) == 2 && getTag(array2[0], n, n + n3) * getTag(array2[1], n, n + n3) <= 0;
        }
        if (n7 == 0) {
            fillEqn(array, (n5 < 0) ? n : (n + n3), x1, ctrlX, x2);
            return solveQuadratic(array, array2) == 2 && evalQuadratic(array2, 2, true, true, null, y1, ctrlY, y2) == 2 && getTag(array2[0], n2, n2 + n4) * getTag(array2[1], n2, n2 + n4) <= 0;
        }
        double n9 = x2 - x1;
        double n10 = y2 - y1;
        double n11 = y2 * x1 - x2 * y1;
        if (tag == 0) {
        }
        else {
            n5 = getTag((n11 + n9 * ((tag < 0) ? n2 : (n2 + n4))) / n10, n, n + n3);
        }
        if (tag2 == 0) {
        }
        else {
            n6 = getTag((n11 + n9 * ((tag2 < 0) ? n2 : (n2 + n4))) / n10, n, n + n3);
        }
        if (n5 * n6 <= 0) {
            return true;
        }
        tag2 = ((n5 * n5 <= 0) ? tag : tag2);
        fillEqn(array, (n6 < 0) ? n : (n + n3), x1, ctrlX, x2);
        evalQuadratic(array2, solveQuadratic(array, array2), true, true, null, y1, ctrlY, y2);
        return tag2 * getTag(array2[0], n2, n2 + n4) <= 0;
    }
    
    @Override
    public boolean intersects(Rectangle2D rectangle2D) {
        return this.intersects(rectangle2D.getX(), rectangle2D.getY(), rectangle2D.getWidth(), rectangle2D.getHeight());
    }
    
    @Override
    public boolean contains(double n, double n2, double n3, double n4) {
        return n3 > 0.0 && n4 > 0.0 && this.contains(n, n2) && this.contains(n + n3, n2) && this.contains(n + n3, n2 + n4) && this.contains(n, n2 + n4);
    }
    
    @Override
    public boolean contains(Rectangle2D rectangle2D) {
        return this.contains(rectangle2D.getX(), rectangle2D.getY(), rectangle2D.getWidth(), rectangle2D.getHeight());
    }
    
    @Override
    public Rectangle getBounds() {
        return this.getBounds2D().getBounds();
    }
    
    @Override
    public PathIterator getPathIterator(AffineTransform affineTransform) {
        return new QuadIterator(this, affineTransform);
    }
    
    @Override
    public PathIterator getPathIterator(AffineTransform affineTransform, double n) {
        return new FlatteningPathIterator(this.getPathIterator(affineTransform), n);
    }
    
    public Object clone() {
        try {
            return super.clone();
        }
        catch (CloneNotSupportedException cause) {
            throw new InternalError(cause);
        }
    }
    
    public static class Float extends QuadCurve2D implements Serializable
    {
        public float x1;
        public float y1;
        public float ctrlx;
        public float ctrly;
        public float x2;
        public float y2;
        private static final long serialVersionUID = -8511188402130719609L;
        
        public Float() {
        }
        
        public Float(float n, float n2, float n3, float n4, float n5, float n6) {
            this.setCurve(n, n2, n3, n4, n5, n6);
        }
        
        @Override
        public double getX1() {
            return (double)this.x1;
        }
        
        @Override
        public double getY1() {
            return (double)this.y1;
        }
        
        @Override
        public Point2D getP1() {
            return new Point2D.Float(this.x1, this.y1);
        }
        
        @Override
        public double getCtrlX() {
            return (double)this.ctrlx;
        }
        
        @Override
        public double getCtrlY() {
            return (double)this.ctrly;
        }
        
        @Override
        public Point2D getCtrlPt() {
            return new Point2D.Float(this.ctrlx, this.ctrly);
        }
        
        @Override
        public double getX2() {
            return (double)this.x2;
        }
        
        @Override
        public double getY2() {
            return (double)this.y2;
        }
        
        @Override
        public Point2D getP2() {
            return new Point2D.Float(this.x2, this.y2);
        }
        
        @Override
        public void setCurve(double n, double n2, double n3, double n4, double n5, double n6) {
            this.x1 = (float)n;
            this.y1 = (float)n2;
            this.ctrlx = (float)n3;
            this.ctrly = (float)n4;
            this.x2 = (float)n5;
            this.y2 = (float)n6;
        }
        
        public void setCurve(float x1, float y1, float ctrlx, float ctrly, float x2, float y2) {
            this.x1 = x1;
            this.y1 = y1;
            this.ctrlx = ctrlx;
            this.ctrly = ctrly;
            this.x2 = x2;
            this.y2 = y2;
        }
        
        @Override
        public Rectangle2D getBounds2D() {
            float min = Math.min(Math.min(this.x1, this.x2), this.ctrlx);
            float min2 = Math.min(Math.min(this.y1, this.y2), this.ctrly);
            return new Rectangle2D.Float(min, min2, Math.max(Math.max(this.x1, this.x2), this.ctrlx) - min, Math.max(Math.max(this.y1, this.y2), this.ctrly) - min2);
        }
    }
    
    public static class Double extends QuadCurve2D implements Serializable
    {
        public double x1;
        public double y1;
        public double ctrlx;
        public double ctrly;
        public double x2;
        public double y2;
        private static final long serialVersionUID = 4217149928428559721L;
        
        public Double() {
        }
        
        public Double(double n, double n2, double n3, double n4, double n5, double n6) {
            this.setCurve(n, n2, n3, n4, n5, n6);
        }
        
        @Override
        public double getX1() {
            return this.x1;
        }
        
        @Override
        public double getY1() {
            return this.y1;
        }
        
        @Override
        public Point2D getP1() {
            return new Point2D.Double(this.x1, this.y1);
        }
        
        @Override
        public double getCtrlX() {
            return this.ctrlx;
        }
        
        @Override
        public double getCtrlY() {
            return this.ctrly;
        }
        
        @Override
        public Point2D getCtrlPt() {
            return new Point2D.Double(this.ctrlx, this.ctrly);
        }
        
        @Override
        public double getX2() {
            return this.x2;
        }
        
        @Override
        public double getY2() {
            return this.y2;
        }
        
        @Override
        public Point2D getP2() {
            return new Point2D.Double(this.x2, this.y2);
        }
        
        @Override
        public void setCurve(double x1, double y1, double ctrlx, double ctrly, double x2, double y2) {
            this.x1 = x1;
            this.y1 = y1;
            this.ctrlx = ctrlx;
            this.ctrly = ctrly;
            this.x2 = x2;
            this.y2 = y2;
        }
        
        @Override
        public Rectangle2D getBounds2D() {
            double min = Math.min(Math.min(this.x1, this.x2), this.ctrlx);
            double min2 = Math.min(Math.min(this.y1, this.y2), this.ctrly);
            return new Rectangle2D.Double(min, min2, Math.max(Math.max(this.x1, this.x2), this.ctrlx) - min, Math.max(Math.max(this.y1, this.y2), this.ctrly) - min2);
        }
    }
}
